package com.nwalsh.saxon6;

import java.util.Stack;
import java.util.StringTokenizer;
import org.xml.sax.*;
import org.w3c.dom.*;
import com.icl.saxon.Builder;
import com.icl.saxon.Context;
import com.icl.saxon.expr.*;
import com.icl.saxon.functions.Extensions;
import com.icl.saxon.om.*;
import com.icl.saxon.output.*;
import com.icl.saxon.pattern.*;
import com.icl.saxon.tinytree.TinyBuilder;
import com.icl.saxon.tree.*;

/**
 * <p>Saxon extension to decorate a result tree fragment with callouts.</p>
 *
 * <p>$Id: CalloutEmitter.java,v 1.2 2000/11/28 23:05:55 ndw Exp $</p>
 *
 * <p>Copyright (C) 2000 Norman Walsh.</p>
 *
 * <p>This class provides the guts of a
 * <a href="http://users.iclway.co.uk/mhkay/saxon/">Saxon 6.*</a>
 * implementation of callouts for verbatim environments. (It is used
 * by the Verbatim class.)</p>
 *
 * <p>The general design is this: the stylesheets construct a result tree
 * fragment for some verbatim environment. The Verbatim class initializes
 * a CalloutEmitter with information about the callouts that should be applied
 * to the verbatim environment in question. Then the result tree fragment
 * is "replayed" through the CalloutEmitter; the CalloutEmitter builds a
 * new result tree fragment from this event stream, decorated with callouts,
 * and that is returned.</p>
 *
 * <p><b>Change Log:</b></p>
 * <dl>
 * <dt>1.0</dt>
 * <dd><p>Initial release.</p></dd>
 * </dl>
 *
 * @see Verbatim
 *
 * @author Norman Walsh
 * <a href="mailto:ndw@nwalsh.com">ndw@nwalsh.com</a>
 *
 * @version $Id: CalloutEmitter.java,v 1.2 2000/11/28 23:05:55 ndw Exp $
 *
 */
public class CalloutEmitter extends CopyEmitter {
  /** A stack for the preserving information about open elements. */
  protected Stack elementStack = null;

  /** Is the next element absolutely the first element in the fragment? */
  protected boolean firstElement = false;

  /** The FO namespace name. */
  protected static String foURI = "http://www.w3.org/1999/XSL/Format";

  /** The path to callout number graphics. */
  protected String graphicsPath = null;

  /** The extension to use for callout number graphics. */
  protected String graphicsExt = null;

  /** The largest callout number that can be represented as a graphic. */
  protected int graphicsMax = 0;

  /** The default column for callouts that specify only a line. */
  protected int defaultColumn = 60;

  /** Is the stylesheet currently running an FO stylesheet? */
  protected boolean foStylesheet = false;

  /** The current line number. */
  private static int lineNumber = 0;

  /** The current column number. */
  private static int colNumber = 0;

  /** The (sorted) array of callouts obtained from the areaspec. */
  private static Callout callout[] = null;

  /** The number of callouts in the callout array. */
  private static int calloutCount = 0;

  /** A pointer used to keep track of our position in the callout array. */
  private static int calloutPos = 0;

  /** <p>Constructor for the CalloutEmitter.</p>
   *
   * @param namePool The name pool to use for constructing elements and attributes.
   * @param graphicsPath The path to callout number graphics.
   * @param graphicsExt The extension for callout number graphics.
   * @param graphicsMax The largest callout number that can be represented as a graphic.
   * @param defaultColumn The default column for callouts.
   * @param foStylesheet Is this an FO stylesheet?
   */
  public CalloutEmitter(NamePool namePool,
			String graphicsPath,
			String graphicsExt,
			int graphicsMax,
			int defaultColumn,
			boolean foStylesheet) {
    super(namePool);
    elementStack = new Stack();
    firstElement = true;

    this.graphicsPath = graphicsPath;
    this.graphicsExt = graphicsExt;
    this.graphicsMax = graphicsMax;
    this.defaultColumn = defaultColumn;
    this.foStylesheet = foStylesheet;
  }

  /**
   * <p>Examine the areaspec and determine the number and position of 
   * callouts.</p>
   *
   * <p>The <code><a href="http://docbook.org/tdg/html/areaspec.html">areaspecNodeSet</a></code>
   * is examined and a sorted list of the callouts is constructed.</p>
   *
   * <p>This data structure is used to augment the result tree fragment
   * with callout bullets.</p>
   *
   * @param areaspecNodeSet The source document &lt;areaspec&gt; element.
   *
   */
  public void setupCallouts (NodeSetIntent areaspecNodeSet) {
    callout = new Callout[10];
    calloutCount = 0;
    calloutPos = 0;
    lineNumber = 1;
    colNumber = 1;

    // First we walk through the areaspec to calculate the position
    // of the callouts
    //  <areaspec>
    //  <areaset id="ex.plco.const" coords="">
    //    <area id="ex.plco.c1" coords="4"/>
    //    <area id="ex.plco.c2" coords="8"/>
    //  </areaset>
    //  <area id="ex.plco.ret" coords="12"/>
    //  <area id="ex.plco.dest" coords="12"/>
    //  </areaspec>
    try {
      int pos = 0;
      int coNum = 0;
      boolean inAreaSet = false;
      NodeInfo areaspec = areaspecNodeSet.getFirst();
      NodeInfo children[] = areaspec.getAllChildNodes();

      for (int count = 0; count < children.length; count++) {
	NodeInfo node = children[count];
	if (node.getNodeType() == NodeInfo.ELEMENT) {
	  if (node.getNodeName().equalsIgnoreCase("areaset")) {
	    coNum++;
	    NodeInfo areas[] = node.getAllChildNodes();
	    for (int acount = 0; acount < areas.length; acount++) {
	      NodeInfo area = areas[acount];
	      if (area.getNodeType() == NodeInfo.ELEMENT) {
		if (area.getNodeName().equalsIgnoreCase("area")) {
		  addCallout(coNum, area, defaultColumn);
		} else {
		  System.out.println("Unexpected element in areaset: "
				     + area.getNodeName());
		}
	      }
	    }
	  } else if (node.getNodeName().equalsIgnoreCase("area")) {
	    coNum++;
	    addCallout(coNum, node, defaultColumn);
	  } else {
	    System.out.println("Unexpected element in areaspec: "
			       + node.getNodeName());
	  }
	}
      }

      // Now sort them
      java.util.Arrays.sort(callout, 0, calloutCount);
    } catch (SAXException e) {
      //nop;
    }
  }

  /** Process characters. */
  public void characters(char[] chars, int start, int len)
    throws org.xml.sax.SAXException {

    // If we hit characters, then there's no first element...
    firstElement = false;

    if (lineNumber == 0) {
      // if there are any text nodes, there's at least one line
      lineNumber++;
      colNumber = 1;
    }

    // Walk through the text node looking for callout positions
    char[] newChars = new char[len];
    int pos = 0;
    for (int count = start; count < start+len; count++) {
      while (calloutPos < calloutCount
	     && callout[calloutPos].getLine() == lineNumber
	     && callout[calloutPos].getColumn() == colNumber) {
	if (pos > 0) {
	  rtf.characters(newChars, 0, pos);
	  pos = 0;
	}
	formatCallout(callout[calloutPos]);
	calloutPos++;
      }

      if (chars[count] == '\n') {
	// What if we need to pad this line?
	while (calloutPos < calloutCount
	       && callout[calloutPos].getLine() == lineNumber
	       && callout[calloutPos].getColumn() > colNumber) {
	  if (pos > 0) {
	    rtf.characters(newChars, 0, pos);
	    pos = 0;
	  }
	  formatPad(callout[calloutPos].getColumn() - colNumber);
	  colNumber = callout[calloutPos].getColumn();
	  while (calloutPos < calloutCount
		 && callout[calloutPos].getLine() == lineNumber
		 && callout[calloutPos].getColumn() == colNumber) {
	    if (pos > 0) {
	      rtf.characters(newChars, 0, pos);
	      pos = 0;
	    }
	    formatCallout(callout[calloutPos]);
	    calloutPos++;
	  }
	}
	lineNumber++;
	colNumber = 1;
      } else {
	colNumber++;
      }
      newChars[pos++] = chars[count];
    }

    if (pos > 0) {
      rtf.characters(newChars, 0, pos);
    }
  }

  /**
   * <p>Add a formatted callout to the result tree fragment.</p>
   *
   * <p>This method examines the callout and adds appropriate markup
   * to the result tree fragment.</p>
   *
   * @param callout The callout to add.
   */
  protected void formatCallout(Callout callout) {
    ElementInfo area = callout.getArea();
    int num = callout.getCallout();
    String userLabel = null;
    String label = "(" + num + ")";

    if (area.getAttributeList().getValue("label") != null) {
      // If this area has a label, use it
      userLabel = area.getAttributeList().getValue("label");
      label = userLabel;
    } else {
      // Otherwise, if its parent is an areaset and it has a label, use that
      ElementInfo parent = (ElementInfo) area.getParentNode();
      if (parent != null
	  && parent.getLocalName().equalsIgnoreCase("areaset")
	  && parent.getAttributeList().getValue("label") != null) {
	userLabel = parent.getAttributeList().getValue("label");
	label = userLabel;
      }
    }

    char chars[] = new char[label.length()];
    for (int count = 0; count < label.length(); count++) {
      chars[count] = label.charAt(count);
    }

    try {
      if (!foStylesheet) {
	int spanName = namePool.allocate("", "", "span");
	AttributeCollection spanAttr = new AttributeCollection(namePool);
	spanAttr.addAttribute("", "", "class", "CDATA", "callout");
	int namespaces[] = new int[1];
	rtf.startElement(spanName, spanAttr, namespaces, 0);
      }

      if (userLabel == null && graphicsPath != null && num <= graphicsMax) {
	int imgName = 0;
	AttributeCollection imgAttr = null;
	int namespaces[] = new int[1];

	if (foStylesheet) {
	  imgName = namePool.allocate("fo", foURI, "external-graphic");
	  imgAttr = new AttributeCollection(namePool);
	  imgAttr.addAttribute("", "", "src", "CDATA",
			       graphicsPath + num + graphicsExt);
	} else {
	  imgName = namePool.allocate("", "", "img");
	  imgAttr = new AttributeCollection(namePool);
	  imgAttr.addAttribute("", "", "src", "CDATA",
			       graphicsPath + num + graphicsExt);
	  imgAttr.addAttribute("", "", "alt", "CDATA", label);
	}

	rtf.startElement(imgName, imgAttr, namespaces, 0);
	rtf.endElement(imgName);
      } else {
	rtf.characters(chars, 0, label.length());
      }

      if (!foStylesheet) {
	int spanName = namePool.allocate("", "", "span");
	rtf.endElement(spanName);
      }
    } catch (SAXException e) {
      System.out.println("SAX Exception in formatCallout");
    }
  }

  /**
   * <p>Add blanks to the result tree fragment.</p>
   *
   * <p>This method adds <tt>numBlanks</tt> to the result tree fragment.
   * It's used to pad lines when callouts occur after the last existing
   * characater in a line.</p>
   *
   * @param numBlanks The number of blanks to add.
   */
  protected void formatPad(int numBlanks) {
    char chars[] = new char[numBlanks];
    for (int count = 0; count < numBlanks; count++) {
      chars[count] = ' ';
    }

    try {
      rtf.characters(chars, 0, numBlanks);
    } catch (SAXException e) {
      System.out.println("SAX Exception in formatCallout");
    }
  }

  /**
   * <p>Add a callout to the global callout array</p>
   *
   * <p>This method examines a callout <tt>area</tt> and adds it to
   * the global callout array if it can be interpreted.</p>
   *
   * <p>Only the <tt>linecolumn</tt> and <tt>linerange</tt> units are
   * supported. If no unit is specifed, <tt>linecolumn</tt> is assumed.
   * If only a line is specified, the callout decoration appears in
   * the <tt>defaultColumn</tt>.</p>
   *
   * @param coNum The callout number.
   * @param node The <tt>area</tt>.
   * @param defaultColumn The default column for callouts.
   */
  protected void addCallout (int coNum,
			     NodeInfo node,
			     int defaultColumn) {
    ElementInfo area = (ElementInfo) node;
    ExtendedAttributes attr = area.getAttributeList();
    String units  = attr.getValue("units");
    String coords = attr.getValue("coords");

    if (units != null
	&& !units.equalsIgnoreCase("linecolumn")
	&& !units.equalsIgnoreCase("linerange")) {
      System.out.println("Only linecolumn and linerange units are supported");
      return;
    }

    if (coords == null) {
      System.out.println("Coords must be specified");
      return;
    }

    // Now let's see if we can interpret the coordinates...
    StringTokenizer st = new StringTokenizer(coords);
    int tokenCount = 0;
    int c1 = 0;
    int c2 = 0;
    while (st.hasMoreTokens()) {
      tokenCount++;
      if (tokenCount > 2) {
	System.out.println("Unparseable coordinates");
	return;
      }
      try {
	String token = st.nextToken();
	int coord = Integer.parseInt(token);
	c2 = coord;
	if (tokenCount == 1) {
	  c1 = coord;
	}
      } catch (NumberFormatException e) {
	System.out.println("Unparseable coordinate");
	return;
      }
    }

    // Make sure we aren't going to blow past the end of our array
    if (calloutCount == callout.length) {
      Callout bigger[] = new Callout[calloutCount+10];
      for (int count = 0; count < callout.length; count++) {
	bigger[count] = callout[count];
      }
      callout = bigger;
    }

    // Ok, add the callout
    if (tokenCount == 2) {
      if (units != null && units.equalsIgnoreCase("linerange")) {
	for (int count = c1; count <= c2; count++) {
	  callout[calloutCount++] = new Callout(coNum, area,
						count, defaultColumn);
	}
      } else {
	// assume linecolumn
	callout[calloutCount++] = new Callout(coNum, area, c1, c2);
      }
    } else {
      // if there's only one number, assume it's the line
      callout[calloutCount++] = new Callout(coNum, area, c1, defaultColumn);
    }
  }

  /** Process end element events. */
  public void endElement(int nameCode)
    throws org.xml.sax.SAXException {
    if (!elementStack.empty()) {
      // if we didn't push the very first element (an fo:block or
      // pre or div surrounding the whole block), then the stack will
      // be empty when we get to the end of the first element...
      elementStack.pop();
    }
    rtf.endElement(nameCode);
  }

  /** Process start element events. */
  public void startElement(int nameCode,
			   org.xml.sax.Attributes attributes,
			   int[] namespaces,
			   int nscount)
    throws org.xml.sax.SAXException {

    if (!skipThisElement(nameCode)) {
      StartElementInfo sei = new StartElementInfo(nameCode, attributes,
						  namespaces, nscount);
      elementStack.push(sei);
    }

    firstElement = false;

    rtf.startElement(nameCode, attributes, namespaces, nscount);
  }

  /**
   * <p>Protect the outer-most block wrapper.</p>
   *
   * <p>Open elements in the result tree fragment are closed and reopened
   * around callouts (so that callouts don't appear inside links or other
   * environments). But if the result tree fragment is a single block
   * (a div or pre in HTML, an fo:block in FO), that outer-most block is
   * treated specially.</p>
   *
   * <p>This method returns true if the element in question is that
   * outermost block.</p>
   *
   * @param nameCode The name code for the element
   *
   * @return True if the element is the outer-most block, false otherwise.
   */
  protected boolean skipThisElement(int nameCode) {
    if (firstElement) {
      int foBlockFingerprint = namePool.getFingerprint(foURI, "block");
      int htmlPreFingerprint = namePool.getFingerprint("", "pre");
      int htmlDivFingerprint = namePool.getFingerprint("", "div");

      if ((foStylesheet && nameCode == foBlockFingerprint)
	  || (!foStylesheet && (nameCode == htmlPreFingerprint
				|| nameCode == htmlDivFingerprint))) {
	// Don't push the outer-most wrapping div, pre, or fo:block
	return true;
      }
    }

    return false;
  }

  /**
   * <p>A private class for maintaining the information required to call
   * the startElement method.</p>
   *
   * <p>In order to close and reopen elements, information about those
   * elements has to be maintained. This class is just the little record
   * that we push on the stack to keep track of that info.</p>
   */
  private class StartElementInfo {
    private int _nameCode;
    org.xml.sax.Attributes _attributes;
    int[] _namespaces;
    int _nscount;

    public StartElementInfo(int nameCode,
			    org.xml.sax.Attributes attributes,
			    int[] namespaces,
			    int nscount) {
      _nameCode = nameCode;
      _attributes = attributes;
      _namespaces = namespaces;
      _nscount = nscount;
    }

    public int getNameCode() {
      return _nameCode;
    }

    public org.xml.sax.Attributes getAttributes() {
      return _attributes;
    }

    public int[] getNamespaces() {
      return _namespaces;
    }

    public int getNSCount() {
      return _nscount;
    }
  }

  /**
   * <p>A private class for maintaining information about callouts.</p>
   *
   * <p>To make processing callouts easier, they are parsed out of the
   * input structure and stored in a sorted array. (The array is sorted
   * according to the order in which the callouts occur.)</p>
   *
   * <p>This class is just the little record
   * that we store in the array for each callout.</p>
   */
  private class Callout implements Comparable {
    /** The callout number. */
    private int callout = 0;
    /** The area ElementInfo item that generated this callout. */
    private ElementInfo area = null;
    /** The line on which this callout occurs. */
    private int line = 0;
    /** The column in which this callout appears. */
    private int col = 0;

    /** The constructor; initialize the private data structures. */
    public Callout(int callout, ElementInfo area, int line, int col) {
      this.callout = callout;
      this.area = area;
      this.line = line;
      this.col = col;
    }

    /**
     * <p>The compareTo method compares this Callout with another.</p>
     *
     * <p>Given two Callouts, A and B, A < B if:</p>
     *
     * <ol>
     * <li>A.line < B.line, or</li>
     * <li>A.line = B.line && A.col < B.col, or</li>
     * <li>A.line = B.line && A.col = B.col && A.callout < B.callout</li>
     * <li>Otherwise, they're equal.</li>
     * </ol>
     */
    public int compareTo (Object o) {
      Callout c = (Callout) o;

      if (line == c.getLine()) {
	if (col > c.getColumn()) {
	  return 1;
	} else if (col < c.getColumn()) {
	  return -1;
	} else {
	  if (callout < c.getCallout()) {
	    return -1;
	  } else if (callout > c.getCallout()) {
	    return 1;
	  } else {
	    return 0;
	  }
	}
      } else {
	if (line > c.getLine()) {
	  return 1;
	} else {
	  return -1;
	}
      }
    }

    /** Access the Callout's area. */
    public ElementInfo getArea() {
      return area;
    }

    /** Access the Callout's line. */
    public int getLine() {
      return line;
    }

    /** Access the Callout's column. */
    public int getColumn() {
      return col;
    }

    /** Access the Callout's callout number. */
    public int getCallout() {
      return callout;
    }
  }
}
